"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getFfiLib = void 0;
var internals_1 = require("./internals");
var logger_1 = require("../logger");
var declarations_1 = require("./declarations");
var LOG_ENV_VAR_NAME = 'PACT_LOG_LEVEL';
var PACT_FFI_VERSION = '0.0.2';
var ffi;
var ffiLogLevel;
var initialiseFfi = function (logLevel) {
    logger_1.default.debug("Initalising native core at log level '" + logLevel + "'");
    process.env[LOG_ENV_VAR_NAME] = logLevel.toUpperCase();
    ffiLogLevel = logLevel;
    var lib = internals_1.createFfi(internals_1.libName('pact_ffi', "v" + PACT_FFI_VERSION), declarations_1.declarations);
    lib.pactffi_init(LOG_ENV_VAR_NAME);
    return lib;
};
var getFfiLib = function (logLevel) {
    if (logLevel === void 0) { logLevel = logger_1.DEFAULT_LOG_LEVEL; }
    if (!ffi) {
        logger_1.default.trace('Initiliasing ffi for the first time');
        ffi = initialiseFfi(logLevel);
    }
    else {
        logger_1.default.trace('Ffi has already been initialised, no need to repeat it');
        if (logLevel !== ffiLogLevel) {
            logger_1.default.warn("The pact native core has already been initialised at log level '" + ffiLogLevel + "'");
            logger_1.default.warn("The new requested log level '" + logLevel + "' will be ignored");
        }
    }
    return ffi;
};
exports.getFfiLib = getFfiLib;
//# sourceMappingURL=index.js.map